#include "net.h"

#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <netdb.h>

int net_client_init(net_t *const n, const char *const raddr, const char *const rport) {
    if (raddr == NULL || rport == NULL)
        return  -1;

    struct addrinfo hints, *ri, *cri;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_DGRAM;

    if (getaddrinfo(raddr, rport, &hints, &ri) != 0)
        return -1;

    for (cri = ri; cri != NULL; cri = cri->ai_next) {
        if ((n->fd = socket(cri->ai_family, cri->ai_socktype, cri->ai_protocol)) == -1)
            continue;

        break;
    }

    if (cri == NULL) {
        freeaddrinfo(ri);
        return -1;
    }

    memset(&n->inaddr, 0, sizeof(n->inaddr));
    n->inaddr_len = sizeof(n->inaddr);

    memset(&n->laddr, 0, sizeof(n->laddr));

    memset(&n->raddr, 0, sizeof(n->raddr));
    memcpy(&n->raddr, cri->ai_addr, sizeof(n->raddr));
    n->raddr_len = sizeof(n->raddr);

    freeaddrinfo(ri);

    return 0;
}

int net_server_init(net_t *const n, const char *const laddr, const char *const lport) {
    struct addrinfo hints, *li, *cli;
    int reuseaddr = 1;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_DGRAM;
    hints.ai_flags = AI_PASSIVE;

    if (getaddrinfo(laddr, lport, &hints, &li) != 0)
        return -1;

    for (cli = li; cli != NULL; cli = cli->ai_next) {
        if ((n->fd = socket(cli->ai_family, cli->ai_socktype, cli->ai_protocol)) == -1)
            continue;

        if (setsockopt(n->fd, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(reuseaddr)) == -1) {
            close(n->fd);
            continue;
        }

        if (bind(n->fd, cli->ai_addr, cli->ai_addrlen) == -1) {
            close(n->fd);
            continue;
        }

        break;
    }

    memset(&n->inaddr, 0, sizeof(n->inaddr));
    n->inaddr_len = sizeof(n->inaddr);

    memset(&n->raddr, 0, sizeof(n->raddr));
    n->raddr_len = sizeof(n->raddr);

    memset(&n->laddr, 0, sizeof(n->laddr));
    memcpy(&n->laddr, cli->ai_addr, sizeof(n->raddr));

    freeaddrinfo(li);

    return 0;
}

void net_destroy(net_t *n) {
    if (n->fd != -1) {
        close(n->fd);
        n->fd = -1;
    }

    memset(&n->raddr, 0, sizeof(n->raddr));
    n->raddr_len = sizeof(n->raddr);
    memset(&n->laddr, 0, sizeof(n->laddr));
    memset(&n->inaddr, 0, sizeof(n->inaddr));
    n->inaddr_len = sizeof(n->inaddr);
}

bool net_set_timeout(net_t *const n, int usec) {
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = usec;

    if (setsockopt(n->fd, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout)) < 0)
        return false;
    if (setsockopt(n->fd, SOL_SOCKET, SO_SNDTIMEO, &timeout, sizeof(timeout)) < 0)
        return false;

    return true;
}

ssize_t net_send(net_t *const n, const char *const buf, int buf_len) {
    if (n->fd == -1 || n->raddr.sa_family == AF_UNSPEC)
        return -1;

    return sendto(n->fd, buf, buf_len, 0, &n->raddr, sizeof(n->raddr));
}

ssize_t net_recv(net_t *const n, void *const buf, size_t buf_len) {
    if (n->fd == -1)
        return -1;

    return recvfrom(n->fd, buf, buf_len, 0, &n->inaddr, &n->inaddr_len);
}
